;; Generated package description from sed-mode.el  -*- no-byte-compile: t -*-
(define-package "sed-mode" "1.0.0.20201201.170236" "Major mode to edit sed scripts" 'nil :url "https://elpa.gnu.org/packages/sed-mode.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
